use base64::{Engine as _, engine::general_purpose};
use std::error::Error;

pub fn encode(text: &String) -> String {
    let mut temp = Vec::with_capacity(text.len());
    for ch in text.bytes() {
        temp.push( ch ^ 'U' as u8 );
    }
    general_purpose::STANDARD.encode(temp)
}

pub fn decode(text: &String) -> Result<String, Box<dyn Error>> {
    let chars = general_purpose::STANDARD.decode(text)?;
    let mut temp = Vec::with_capacity(chars.len());
    for ch in chars {
        temp.push( ch ^ 'U' as u8 );
    }
    Ok(String::from_utf8(temp)?)
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn encode_test() {
        let t1: String = String::from("hello world");
        let t2 = encode(&t1);
        assert_eq!(t2, "PTA5OTp1IjonOTE=");
    }

    #[test]
    fn encode_utf8_test() {
        let t1: String = String::from("🍺");
        let t2 = encode(&t1);
        assert_eq!(t2, "pcrY7w==");
    }

    #[test]
    fn decode_test() {
        let t1: String = String::from("ICYiPCE2PW9vMTA2OjEw");
        let t2 = decode(&t1);
        match t2 {
            Ok(s) => assert_eq!(s, "uswitch::decode"),
            Err(err) => panic!("{}", err),
        }
    }

    #[test]
    fn decode_utf8_test() {
        let t1: String = String::from("t9f5");
        let t2 = decode(&t1);
        match t2 {
            Ok(s) => assert_eq!(s, "€"),
            Err(err) => panic!("{}", err),
        }
    }

    #[test]
    fn decode_fail_test() {
        let t1: String = String::from("yAyL3WlysPBdTEnPDs+JnWlytjv"); // Invalid base64
        let t2 = decode(&t1);
        match t2 {
            Ok(s) => panic!("Expected failure! Got '{}'", s),
            Err(_) => ()
        }
    }

    #[test]
    fn decode_fail_utf8_test() {
        let t1: String = String::from("woXChQ=="); // Invalid UTF-8
        let t2 = decode(&t1);
        match t2 {
            Ok(s) => panic!("Expected failure! Got '{}'", s),
            Err(_) => ()
        }
    }
}
